-- to do
-- save channel info and preview components with quickslot for easier retrieval
-- allow editing of quickslot titles by right clicking


import "GaranStuff.Rainbow.Class";
import "GaranStuff.Rainbow.VindarPatch";
-- This is the Load/Save patch which accounts for European formats, written by Vindar
-- load resource strings - this MUST occur before any windows that use strings get loaded or there will be no defaults
import "GaranStuff.Rainbow.Strings";
import "GaranStuff.Rainbow.Table";
import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "GaranStuff.Rainbow.DropDownList";
import "GaranStuff.Rainbow.ColorPicker";
import "GaranStuff.Rainbow.QuickSlotWindow";
import "GaranStuff.Rainbow.OptionsWindow";
import "GaranStuff.Rainbow.FontMetric";

language=1;
debug=false;
debugSetting=PatchDataLoad( Turbine.DataScope.Account, "RBC_Debug");
if debugSetting~=nil then
	debug=debugSetting;
end

colorGold=Turbine.UI.Color(.80,.60,.1);
colorSilver=Turbine.UI.Color(.75,.75,.75);
colorDarkGrey=Turbine.UI.Color(.1,.1,.1);

trimColor=Turbine.UI.Color(.2,.5,.9);
fontColor=Turbine.UI.Color(1,.9,.5);
backColor=Turbine.UI.Color(.05,.05,.05);

fontFace=Turbine.UI.Lotro.Font.Verdana16;

euroFormat=(tonumber("1,000")==1);

string.split=function(str,separator)
	local tmpArray={};
	local tmpElem;
	str=tostring(str);
	if string.find("^$()%.[]*+-?",separator,1,true)~=nil then separator="%"..separator end
	for tmpElem in string.gmatch(str,"[^"..separator.."]*") do
		table.insert(tmpArray,tmpElem);
	end
	local tmpIndex=1;
	while tmpIndex<=#tmpArray do
		if tmpIndex==#tmpArray and tmpArray[tmpIndex]=="" then
			table.remove(tmpArray,tmpIndex);
		else
			if tmpArray[tmpIndex]=="" and tmpArray[tmpIndex+1]~="" then
				table.remove(tmpArray,tmpIndex);
			else
				tmpIndex=tmpIndex+1;
			end
		end
	end
	return tmpArray;
end
string.ltrim=function(str)
	return (string.gsub(str, "^%s*(.-)", "%1"))
end
string.rtrim=function(str)
	return (string.gsub(str, "^(.-)%s*$", "%1"))
end
string.trim=function(str)
	return (string.gsub(str, "^%s*(.-)%s*$", "%1"))
end

function UnloadPlugin()
	SaveData();
end

function SaveData()

	PatchDataSave( Turbine.DataScope.Account, "RBC_Debug", debug );
	local settings={};
	settings[#settings+1]={"Top",mainWindow:GetTop()};
	settings[#settings+1]={"Left",mainWindow:GetLeft()};
	settings[#settings+1]={"Height",mainWindow:GetHeight()};
	settings[#settings+1]={"Width",mainWindow:GetWidth()};
	settings[#settings+1]={"Opacity",mainWindow.Opacity};
	settings[#settings+1]={"Version",Plugins["Rainbow"]:GetVersion()};

	settings[#settings+1]={"QSTop",qsWindow:GetTop()};
	settings[#settings+1]={"QSLeft",qsWindow:GetLeft()};

	settings[#settings+1]={"fontColor",fontColor};
	settings[#settings+1]={"trimColor",trimColor};
	settings[#settings+1]={"backColor",backColor};
	settings[#settings+1]={"language",language};
	settings[#settings+1]={"euroFormat",euroFormat};

	PatchDataSave( Turbine.DataScope.Account, "RainbowSettings", settings);

	settings={};
	local slot;
	for slot=1,#qsWindow.QSlot do
		settings[slot]={qsWindow.QSlot[slot].Cmd:GetShortcut():GetData(),tostring(qsWindow.QSlot[slot].Channel),qsWindow.QSlot[slot].Clips};
	end
	PatchDataSave( Turbine.DataScope.Character, "RainbowQuickSlots", settings);
end

function LoadQuickSlots()
	local tmpSettings;
	local tmpSlot;
	local tmpString;
	tmpSettings=PatchDataLoad(Turbine.DataScope.Character, "RainbowQuickSlots");
	if (tmpSettings~=nil) then
		local priorFormat=euroFormat;
		if tostring(type(tmpSettings))=="table" then
			for index=1,#tmpSettings do
				if tmpSettings[index]~=nil then
					tmpSlot=#qsWindow.QSlot+1;
					qsWindow.QSlot[tmpSlot]=Turbine.UI.Control();
					qsWindow.QSlot[tmpSlot]:SetParent(qsWindow);
					qsWindow.QSlot[tmpSlot]:SetSize(32,20);
					qsWindow.QSlot[tmpSlot]:SetPosition(0,(tmpSlot-1)*20+16);
					qsWindow.QSlot[tmpSlot].Cmd=Turbine.UI.Lotro.Quickslot();
					qsWindow.QSlot[tmpSlot].Cmd:SetParent(qsWindow.QSlot[tmpSlot]);
					qsWindow.QSlot[tmpSlot].Cmd:SetSize(32,52);
					qsWindow.QSlot[tmpSlot].Cmd:SetPosition(0,-32);
					qsWindow.QSlot[tmpSlot].Clips={};
					if type(tmpSettings[index])=="table" then
						qsWindow.QSlot[tmpSlot].Cmd:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, tmpSettings[index][1]));
						qsWindow.QSlot[tmpSlot].Channel=tmpSettings[index][2]
						for tmpIndex=1,#tmpSettings[index][3] do
							qsWindow.QSlot[tmpSlot].Clips[tmpIndex]={tmpSettings[index][3][tmpIndex][1],tmpSettings[index][3][tmpIndex][2],tmpSettings[index][3][tmpIndex][3]};
						end
					else
						-- old version shortcuts
						qsWindow.QSlot[tmpSlot].Cmd:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, tmpSettings[index]));
						qsWindow.QSlot[tmpSlot].Channel="";
					end

					qsWindow.QSlot[tmpSlot].Label=Turbine.UI.Label();
					qsWindow.QSlot[tmpSlot].Label:SetParent(qsWindow);
					qsWindow.QSlot[tmpSlot].Label:SetSize(32,20);
					qsWindow.QSlot[tmpSlot].Label:SetPosition(0,(tmpSlot-1)*20+16);
					qsWindow.QSlot[tmpSlot].Label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
					qsWindow.QSlot[tmpSlot].Label:SetText(tostring(tmpSlot));
					qsWindow.QSlot[tmpSlot].Label:SetMouseVisible(false);
				end
			end
			qsWindow:SetHeight(#tmpSettings*20+16);
		end
	end
end


RainbowWindow = class( Turbine.UI.Lotro.Window );

function RainbowWindow:Constructor()
	Turbine.UI.Lotro.Window.Constructor( self );
	self:SetSize(640,375);
	self.Loaded=false;

	-- load the settings and set the window position if possible, otherwise default to center
	local initHeight=375;
	local initWidth=640;
	local initLeft=(Turbine.UI.Display:GetWidth() - initWidth) / 2;
	local initTop=(Turbine.UI.Display:GetHeight() - initHeight) / 2;

	self.Opacity=1;

	self.SelectedClip=0;
	self.Resizing=false;
	self.SayText="";
	self.MoveX=-1;
	self.MoveY=-1;
	self.LoadMinimized=true;
	self.ShowIconWhenMinimized=true;
	tmpSettings=PatchDataLoad(Turbine.DataScope.Account, "RainbowSettings");
	if (tmpSettings~=nil) then
		local priorFormat=euroFormat;
		if tostring(type(tmpSettings))=="table" then
			for index=1,#tmpSettings do
				if tostring(type(tmpSettings[index]))=="table" then
					if (tmpSettings[index][1]=="Top") then
						initTop=tmpSettings[index][2];
					elseif (tmpSettings[index][1]=="Left") then
						initLeft=tmpSettings[index][2];
					elseif (tmpSettings[index][1]=="Height") then
						initHeight=tmpSettings[index][2];
					elseif (tmpSettings[index][1]=="Width") then
						initWidth=tmpSettings[index][2];
					elseif (tmpSettings[index][1]=="Opacity") then
						self.Opacity=tmpSettings[index][2];
					elseif (tmpSettings[index][1]=="trimColor") then
						if euroFormat then
							-- replace "." with "," tonumber(string.gsub(tmpSettings[index][2].R,".",","))
							trimColor=Turbine.UI.Color(tonumber((string.gsub(tmpSettings[index][2].R,"%.",","))),tonumber((string.gsub(tmpSettings[index][2].G,"%.",","))),tonumber((string.gsub(tmpSettings[index][2].B,"%.",","))));
						else
							trimColor=Turbine.UI.Color(tonumber((string.gsub(tmpSettings[index][2].R,",","."))),tonumber((string.gsub(tmpSettings[index][2].G,",","."))),tonumber((string.gsub(tmpSettings[index][2].B,",","."))));
							-- replace "." with ","
						end
					elseif (tmpSettings[index][1]=="fontColor") then
						if euroFormat then
							-- replace "." with "," tonumber(string.gsub(tmpSettings[index][2].R,".",","))
							fontColor=Turbine.UI.Color(tonumber((string.gsub(tmpSettings[index][2].R,"%.",","))),tonumber((string.gsub(tmpSettings[index][2].G,"%.",","))),tonumber((string.gsub(tmpSettings[index][2].B,"%.",","))));
						else
							fontColor=Turbine.UI.Color(tonumber((string.gsub(tmpSettings[index][2].R,",","."))),tonumber((string.gsub(tmpSettings[index][2].G,",","."))),tonumber((string.gsub(tmpSettings[index][2].B,",","."))));
							-- replace "." with ","
						end
					elseif (tmpSettings[index][1]=="backColor") then
						if euroFormat then
							-- replace "." with "," tonumber(string.gsub(tmpSettings[index][2].R,".",","))
							backColor=Turbine.UI.Color(tonumber((string.gsub(tmpSettings[index][2].R,"%.",","))),tonumber((string.gsub(tmpSettings[index][2].G,"%.",","))),tonumber((string.gsub(tmpSettings[index][2].B,"%.",","))));
						else
							backColor=Turbine.UI.Color(tonumber((string.gsub(tmpSettings[index][2].R,",","."))),tonumber((string.gsub(tmpSettings[index][2].G,",","."))),tonumber((string.gsub(tmpSettings[index][2].B,",","."))));
							-- replace "." with ","
						end
					end
				end
			end
		else
			-- data was unreadable, we need to display a message to the user that the data could not be read
			dataLoadError=true
		end

		-- account for changed resolution
		if initLeft>(Turbine.UI.Display.GetWidth()-initWidth) then initLeft=Turbine.UI.Display.GetWidth()-initWidth end;
		if initLeft<0 then initLeft=0 end;
		if initWidth>(Turbine.UI.Display.GetWidth()-initLeft) then initWidth=Turbine.UI.Display.GetWidth()-initLeft end;
		if initTop>(Turbine.UI.Display.GetHeight()-initHeight) then initTop=Turbine.UI.Display.GetHeight()-initHeight end;
		if initTop<0 then initTop=0 end;
		if initHeight>(Turbine.UI.Display.GetHeight()-initTop) then initTop=Turbine.UI.Display.GetHeight()-initHeight end;
	end

self.Opacity=1
--	self:SetOpacity(self.Opacity);
--	trimColor.A=Opacity;
--	fontColor.A=Opacity;
--	backColor.A=Opacity;

	self:SetPosition( initLeft, initTop );
--	self:SetHeight(initHeight);
--	self:SetWidth(initWidth);
	self:SetVisible(false);
	self:SetText(Resource[language][2]);

	self.TextCaption=Turbine.UI.Label();
	self.TextCaption:SetParent(self);
	self.TextCaption:SetSize(100,20);
	self.TextCaption:SetPosition(20,45);
--	self.TextCaption:SetBackColor(backColor);
	self.TextCaption:SetForeColor(fontColor);
	self.TextCaption:SetFont(fontFace);
	self.TextCaption:SetText(Resource[language][24]..":");

	self.TextEntry=Turbine.UI.Lotro.TextBox();
	self.TextEntry:SetParent(self);
	self.TextEntry:SetSize(self:GetWidth()-40-self.TextCaption:GetWidth(),60);
	self.TextEntry:SetPosition(self.TextCaption:GetLeft()+self.TextCaption:GetWidth(),self.TextCaption:GetTop());
	self.TextEntry:SetBackColor(backColor);
	self.TextEntry:SetForeColor(fontColor);
	self.TextEntry:SetFont(fontFace);

	self.fontMetric=FontMetric();
	self.fontMetric:SetFont(Turbine.UI.Lotro.Font.Verdana12); -- use the default chat font

	self.ColorCaption=Turbine.UI.Label();
	self.ColorCaption:SetParent(self);
	self.ColorCaption:SetSize(100,20);
	self.ColorCaption:SetPosition(self.TextCaption:GetLeft(),self.TextEntry:GetTop()+self.TextEntry:GetHeight()+15);
--	self.ColorCaption:SetBackColor(backColor);
	self.ColorCaption:SetForeColor(fontColor);
	self.ColorCaption:SetFont(fontFace);
	self.ColorCaption:SetText(Resource[language][22]..":");

	self.Color=ColorPicker();
	self.Color:SetParent(self);
	self.Color:SetColor(fontColor);
	self.Color:SetBackColor(backColor);
	self.Color:SetTrimColor(trimColor);
	self.Color:SetWidth(200);
	self.Color:SetHeight(29);
	self.Color:SetPosition(self.ColorCaption:GetLeft()+self.ColorCaption:GetWidth()+10,self.ColorCaption:GetTop())

	self.CBNoColor=Turbine.UI.Lotro.CheckBox();
	self.CBNoColor:SetParent(self);
	self.CBNoColor:SetForeColor(fontColor);
	self.CBNoColor:SetPosition(self.Color:GetLeft()+self.Color:GetWidth()+20,self.ColorCaption:GetTop())
	self.CBNoColor:SetSize(self:GetWidth()-self.CBNoColor:GetLeft()-10,18);
	self.CBNoColor:SetBackColor(backColor);
	self.CBNoColor:SetText(Resource[language][25]);
	self.CBNoColor.CheckedChanged=function()
		self.Color:SetVisible(not self.CBNoColor:IsChecked());
	end
	self.CBNoColor:SetChecked(false);

	self.ClearButton=Turbine.UI.Lotro.Button();
	self.ClearButton:SetParent(self);
	self.ClearButton:SetSize(100,20);
	self.ClearButton:SetPosition(self:GetWidth()-self.ColorCaption:GetLeft()-self.ClearButton:GetWidth(),self.Color:GetTop()+self.Color:GetHeight()+15)
	self.ClearButton:SetFont(fontFace);
	self.ClearButton:SetForeColor(fontColor);
	self.ClearButton:SetText(Resource[language][16]);
	self.ClearButton.Click=function(sender,args)
		local tmpIndex=#self.Preview.Clips
		while tmpIndex>=1 do
			self.Preview.Clips[tmpIndex].MouseClick=nil;
			self.Preview.Clips[tmpIndex]:SetVisible(false);
			self.Preview.Clips[tmpIndex]=nil;
			table.remove(self.Preview.Clips,tmpIndex);
			tmpIndex=tmpIndex-1;
		end
		self.SayText="";
		self:SetQS();
		self.SelectedClip=0;
	end

	self.AddButton=Turbine.UI.Lotro.Button();
	self.AddButton:SetParent(self);
	self.AddButton:SetSize(100,20);
	self.AddButton:SetPosition(self.Color:GetLeft(),self.ClearButton:GetTop())
	self.AddButton:SetFont(fontFace);
	self.AddButton:SetForeColor(fontColor);
	self.AddButton:SetText(Resource[language][15]);
	self.AddButton.Click=function(sender,args)
		self:AddText(self.TextEntry:GetText(),self.Color:GetColor(),self.CBNoColor:IsChecked());
	end

	self.ReplaceButton=Turbine.UI.Lotro.Button();
	self.ReplaceButton:SetParent(self);
	self.ReplaceButton:SetSize(100,20);
	self.ReplaceButton:SetPosition((self.ClearButton:GetLeft()-self.AddButton:GetLeft()-self.AddButton:GetWidth()-self.ReplaceButton:GetWidth())/2+self.AddButton:GetLeft()+self.AddButton:GetWidth(),self.ClearButton:GetTop())
	self.ReplaceButton:SetFont(fontFace);
	self.ReplaceButton:SetForeColor(fontColor);
	self.ReplaceButton:SetText(Resource[language][19]);
	self.ReplaceButton.Click=function(sender,args)
		if self.SelectedClip~=nil and self.SelectedClip~=0 then
			self:ReplaceText(self.TextEntry:GetText(),self.Color:GetColor());
		end
	end

	self.Sep1=Turbine.UI.Control();
	self.Sep1:SetParent(self);
	self.Sep1:SetSize(self:GetWidth()-40,1);
	self.Sep1:SetPosition(20,self.ReplaceButton:GetTop()+self.ReplaceButton:GetHeight()+10);
	self.Sep1:SetBackColor(trimColor);

	self.ChannelCaption=Turbine.UI.Label();
	self.ChannelCaption:SetParent(self);
	self.ChannelCaption:SetSize(100,20);
	self.ChannelCaption:SetPosition(self.ColorCaption:GetLeft(),self.ReplaceButton:GetTop()+self.ReplaceButton:GetHeight()+20);
--	self.ChannelCaption:SetBackColor(backColor);
	self.ChannelCaption:SetForeColor(fontColor);
	self.ChannelCaption:SetFont(fontFace);
	self.ChannelCaption:SetText(Resource[language][20]..":");

	self.ChannelList=DropDownList()
	self.ChannelList:SetParent(self);
	self.ChannelList:SetPosition(self.AddButton:GetLeft(),self.ReplaceButton:GetTop()+self.ReplaceButton:GetHeight()+20);
	self.ChannelList:SetSize(100,20);
	self.ChannelList:SetBackColor(backColor);
	self.ChannelList:SetTextColor(fontColor);
	self.ChannelList:SetCurrentBackColor(backColor);
	self.ChannelList:SetDropRows(7);

	self.ChannelIcon=Turbine.UI.Control();
	self.ChannelIcon:SetParent(self);
	self.ChannelIcon:SetPosition(self.ReplaceButton:GetLeft()-16,self.ChannelList:GetTop()+2);
	self.ChannelIcon:SetSize(16,16);
	self.ChannelIcon:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.ChannelIcon:SetBackground(0x41007e1d);
	self.ChannelIcon:SetVisible(false);

	self.ChannelOther=Turbine.UI.Lotro.TextBox()
	self.ChannelOther:SetParent(self);
	self.ChannelOther:SetPosition(self.ReplaceButton:GetLeft(),self.ChannelList:GetTop());
	self.ChannelOther:SetSize(120,20);
	self.ChannelOther:SetFont(fontFace);
	self.ChannelOther:SetBackColor(backColor);
	self.ChannelOther:SetForeColor(fontColor);
	self.ChannelOther:SetText("");
	self.ChannelOther:SetVisible(false);
	self.ChannelOther.OldValue="";
	self.ChannelOther.FocusGained=function()
		self.ChannelOther:SetWantsUpdates(true);
	end
	self.ChannelOther.FocusLost=function()
		self.ChannelOther:SetWantsUpdates(false);
	end
	self.ChannelOther.Update=function()
		local newValue=self.ChannelOther:GetText();
		if newValue~=self.ChannelOther.OldValue then
			self.ChannelOther.OldValue=newValue;
			self:SetQS();
		end
	end

-- standard chat channels
	self.ChannelList:AddItem("Say","/say"); -- Crier, /crier ?? Parley, /parley
	self.ChannelList:AddItem("Fellowship","/f");
--	if Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment()==Turbine.Gameplay.Alignment.FreePeople then
	if Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment()~=Turbine.Gameplay.Alignment.MonsterPlayer then
		-- not sure if this will work during session play or not...
		self.ChannelList:AddItem("Kinship","/k");
	end
	if Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment()==Turbine.Gameplay.Alignment.MonsterPlayer then
		self.ChannelList:AddItem("Tribe","/tr");
	end
	self.ChannelList:AddItem("AFK","/afk"); -- Absent, /abs

-- doesn't work with raidshout :(

	self.ChannelList:AddItem("Advice","/advice");-- conseil, /conseil
	self.ChannelList:AddItem("Emote","/emote"); -- Emote, /emote
	self.ChannelList:AddItem("LFF","/lff");
	self.ChannelList:AddItem("OOC","/ooc");
	self.ChannelList:AddItem("Officer","/o"); -- Officier,/o
	self.ChannelList:AddItem("RP","/rpc");
	self.ChannelList:AddItem("Raid","/ra"); -- raid, /ra
	self.ChannelList:AddItem("Regional","/regional"); --Rgional, /rgional
	self.ChannelList:AddItem("Trade","/trade"); --changer, /changer
	self.ChannelList:AddItem("UC1","/1"); -- CD1,/1
	self.ChannelList:AddItem("UC2","/2"); -- CD2,/2
	self.ChannelList:AddItem("UC3","/3"); -- CD3,/3
	self.ChannelList:AddItem("UC4","/4"); -- CD4,/4
	self.ChannelList:AddItem("Other","");
	self.ChannelList.SelectedIndexChanged=function()
		self.ChannelIcon:SetVisible(self.ChannelList:GetValue()=="")
		self.ChannelOther:SetVisible(self.ChannelList:GetValue()=="")
		self:SetQS();
	end
-- Tell, /tell -- Dire, /dire
-- Reply /r --rpondre, /rpondre
-- Retell, /retell -- Redire, /rd

-- french chat commands - raccourci=shortcut
-- CJDR ? 
-- comm (communaut) ? fellowship?
-- conf ? kinship
-- rdc ? rpc

	self.QS=Turbine.UI.Lotro.Quickslot();
	self.QS:SetParent(self);
	self.QS:SetSize(96,18);
	self.QS:SetPosition(self.ClearButton:GetLeft()+2,self.ChannelList:GetTop()+2);
	self.QS:SetAllowDrop(false);

	self.QSButton=Turbine.UI.Label();
	self.QSButton:SetParent(self);
	self.QSButton:SetSize(96,18);
	self.QSButton:SetPosition(self.ClearButton:GetLeft()+2,self.ChannelList:GetTop()+2);
	self.QSButton:SetBackground("GaranStuff/Rainbow/Resources/Send.jpg");
	self.QSButton:SetForeColor(fontColor);
	self.QSButton:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
	self.QSButton:SetText(Resource[language][17]);
	self.QSButton:SetMouseVisible(false);
	self.QS.MouseEnter=function()
		self.QSButton:SetBackground("GaranStuff/Rainbow/Resources/Send_highlight.jpg");
	end
	self.QS.MouseLeave=function()
		self.QSButton:SetBackground("GaranStuff/Rainbow/Resources/Send.jpg");
	end

	self.Sep2=Turbine.UI.Control();
	self.Sep2:SetParent(self);
	self.Sep2:SetSize(self:GetWidth()-40,1);
	self.Sep2:SetPosition(20,self.ChannelCaption:GetTop()+self.ChannelCaption:GetHeight()+10);
	self.Sep2:SetBackColor(trimColor);

	self.SlotCaption=Turbine.UI.Label();
	self.SlotCaption:SetParent(self);
	self.SlotCaption:SetSize(100,20);
	self.SlotCaption:SetPosition(self.ColorCaption:GetLeft(),self.ChannelCaption:GetTop()+self.ChannelCaption:GetHeight()+20);
--	self.SlotCaption:SetBackColor(backColor);
	self.SlotCaption:SetForeColor(fontColor);
	self.SlotCaption:SetFont(fontFace);
	self.SlotCaption:SetText(Resource[language][21]..":");

	self.SlotList=DropDownList()
	self.SlotList:SetParent(self);
	self.SlotList:SetPosition(self.AddButton:GetLeft(),self.ChannelList:GetTop()+self.ChannelList:GetHeight()+20);
	self.SlotList:SetSize(100,20);
	self.SlotList:SetBackColor(backColor);
	self.SlotList:SetTextColor(fontColor);
	self.SlotList:SetCurrentBackColor(backColor);
	self.SlotList:SetDropRows(5);

	self.SlotList:AddItem("New",0);
	local slot;
	for slot=1,#qsWindow.QSlot do
		self.SlotList:AddItem("Slot "..tostring(slot),slot);
	end

	self.SlotList.SelectedIndexChanged=function()
	end

	self.AssignButton=Turbine.UI.Lotro.Button();
	self.AssignButton:SetParent(self);
	self.AssignButton:SetSize(100,20);
	self.AssignButton:SetPosition(self.ReplaceButton:GetLeft(),self.SlotList:GetTop());
	self.AssignButton:SetFont(fontFace);
	self.AssignButton:SetForeColor(fontColor);
	self.AssignButton:SetText(Resource[language][18]);
	self.AssignButton.Click=function(sender,args)
-- assign the current text to the currently selected slot
		local tmpSlot=self.SlotList:GetValue();
		if tmpSlot==0 then
			tmpSlot=#qsWindow.QSlot+1;
			qsWindow.QSlot[tmpSlot]=Turbine.UI.Control();
			qsWindow.QSlot[tmpSlot]:SetParent(qsWindow);
			qsWindow.QSlot[tmpSlot]:SetSize(32,20);
			qsWindow.QSlot[tmpSlot]:SetPosition(0,(tmpSlot-1)*20+16);
			qsWindow.QSlot[tmpSlot].Cmd=Turbine.UI.Lotro.Quickslot();
			qsWindow.QSlot[tmpSlot].Cmd:SetParent(qsWindow.QSlot[tmpSlot]);
			qsWindow.QSlot[tmpSlot].Cmd:SetSize(32,52);
			qsWindow.QSlot[tmpSlot].Cmd:SetPosition(0,-32);
			self.SlotList:AddItem("Slot "..tostring(tmpSlot),tmpSlot);
			qsWindow:SetHeight(tmpSlot*20+16);
			qsWindow.QSlot[tmpSlot].Label=Turbine.UI.Label();
			qsWindow.QSlot[tmpSlot].Label:SetParent(qsWindow);
			qsWindow.QSlot[tmpSlot].Label:SetSize(32,20);
			qsWindow.QSlot[tmpSlot].Label:SetPosition(0,(tmpSlot-1)*20+16);
			qsWindow.QSlot[tmpSlot].Label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
			qsWindow.QSlot[tmpSlot].Label:SetText(tostring(tmpSlot));
			qsWindow.QSlot[tmpSlot].Label:SetMouseVisible(false);
		end
		if self.ChannelOther:IsVisible() then
			qsWindow.QSlot[tmpSlot].Cmd:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self.ChannelOther:GetText().." "..self.SayText));
			qsWindow.QSlot[tmpSlot].Channel=self.ChannelOther:GetText()
		else
			qsWindow.QSlot[tmpSlot].Cmd:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self.ChannelList:GetValue().." "..self.SayText));
			qsWindow.QSlot[tmpSlot].Channel=self.ChannelList:GetValue()
		end
		local tmpIndex;
		qsWindow.QSlot[tmpSlot].Clips={};
		for tmpIndex=1, #self.Preview.Clips do
			qsWindow.QSlot[tmpSlot].Clips[tmpIndex]={self.Preview.Clips[tmpIndex]:GetForeColor(),self.Preview.Clips[tmpIndex].NoColor,self.Preview.Clips[tmpIndex]:GetText()};
		end
	end

	self.RecallButton=Turbine.UI.Lotro.Button();
	self.RecallButton:SetParent(self);
	self.RecallButton:SetSize(90,20);
	self.RecallButton:SetPosition(self.AssignButton:GetLeft()+self.AssignButton:GetWidth(),self.SlotList:GetTop());
	self.RecallButton:SetFont(fontFace);
	self.RecallButton:SetForeColor(fontColor);
	self.RecallButton:SetText(Resource[language][26]);
	self.RecallButton.Click=function(sender,args)
		-- parse the quickslot, looking for <rgb tags
		local tmpSlot=self.SlotList:GetValue();
		if tmpSlot>0 then
			local tmpChannel=qsWindow.QSlot[tmpSlot].Channel;
			local tmpOtherIndex=0;
			local tmpChannelIndex=1;
			local found=false;
			if tmpChannel~=nil and tmpChannel~="" then
				-- try to find the correct channel
				
				while tmpChannelIndex<=self.ChannelList.ListData:GetItemCount() and not found do
					if self.ChannelList.ListData:GetItem(tmpChannelIndex):GetText()=="Other" then
						tmpOtherIndex=tmpChannelIndex;
					elseif self.ChannelList.ListData:GetItem(tmpChannelIndex).DataValue==tmpChannel then
						self.ChannelList:SetSelectedIndex(tmpChannelIndex);
						self.ChannelOther:SetVisible(false);
						found=true;
					end
					tmpChannelIndex=tmpChannelIndex+1;
				end
				if not found then
					self.ChannelList:SetSelectedIndex(tmpOtherIndex);
					self.ChannelOther:SetVisible(true);
					self.ChannelOther:SetText(tmpChannel);
				end
				-- clear the quickslot and SayText
				local tmpIndex=#self.Preview.Clips
				while tmpIndex>=1 do
					self.Preview.Clips[tmpIndex].MouseClick=nil;
					self.Preview.Clips[tmpIndex]:SetVisible(false);
					self.Preview.Clips[tmpIndex]=nil;
					table.remove(self.Preview.Clips,tmpIndex);
					tmpIndex=tmpIndex-1;
				end
				self.SayText="";
				self:SetQS();
				self.SelectedClip=0;
				-- now rebuild the Preview
				local tmpIndex;
				for tmpIndex=1,#qsWindow.QSlot[tmpSlot].Clips do
					self:AddText(qsWindow.QSlot[tmpSlot].Clips[tmpIndex][3],Turbine.UI.Color(qsWindow.QSlot[tmpSlot].Clips[tmpIndex][1].A,qsWindow.QSlot[tmpSlot].Clips[tmpIndex][1].R,qsWindow.QSlot[tmpSlot].Clips[tmpIndex][1].G,qsWindow.QSlot[tmpSlot].Clips[tmpIndex][1].B),qsWindow.QSlot[tmpSlot].Clips[tmpIndex][2])
				end
			else
				-- tell user that the quickslot was saved in a prior version and can not be recalled
			end
		end
	end

	self.RemoveButton=Turbine.UI.Lotro.Button();
	self.RemoveButton:SetParent(self);
	self.RemoveButton:SetSize(100,20);
	self.RemoveButton:SetPosition(self.ClearButton:GetLeft(),self.SlotList:GetTop());
	self.RemoveButton:SetFont(fontFace);
	self.RemoveButton:SetForeColor(fontColor);
	self.RemoveButton:SetText(Resource[language][23]);
	self.RemoveButton.Click=function(sender,args)
		-- remove the Quickslot
		local tmpSlot=self.SlotList:GetValue();
		if tmpSlot~=0 then
			qsWindow.QSlot[tmpSlot].Cmd:SetVisible(false);
			qsWindow.QSlot[tmpSlot].Cmd=nil;
			qsWindow.QSlot[tmpSlot].Label:SetVisible(false);
			qsWindow.QSlot[tmpSlot].Label=nil;
			qsWindow.QSlot[tmpSlot]:SetVisible(false);
			qsWindow.QSlot[tmpSlot]:SetParent(nil);
			qsWindow.QSlot[tmpSlot]=nil;
			table.remove(qsWindow.QSlot,tmpSlot);
			self.SlotList:RemoveItemAt(self.SlotList.ListData:GetItemCount())
			for tmpSlot=tmpSlot,#qsWindow.QSlot do
				qsWindow.QSlot[tmpSlot]:SetTop((tmpSlot-1)*20+16);
				qsWindow.QSlot[tmpSlot].Label:SetTop((tmpSlot-1)*20+16);
				qsWindow.QSlot[tmpSlot].Label:SetText(tostring(tmpSlot));
			end
			qsWindow:SetHeight(#qsWindow.QSlot*20+16);
		end
	end

	self.Sep3=Turbine.UI.Control();
	self.Sep3:SetParent(self);
	self.Sep3:SetSize(self:GetWidth()-40,1);
	self.Sep3:SetPosition(20,self.SlotList:GetTop()+self.SlotList:GetHeight()+10);
	self.Sep3:SetBackColor(trimColor);

	self.PreviewCaption=Turbine.UI.Label();
	self.PreviewCaption:SetParent(self);
	self.PreviewCaption:SetSize(100,20);
	self.PreviewCaption:SetPosition(20,self.SlotList:GetTop()+self.SlotList:GetHeight()+20);
	self.PreviewCaption:SetForeColor(fontColor);
	self.PreviewCaption:SetFont(fontFace);
	self.PreviewCaption:SetText(Resource[language][14]..":");
	self.PreviewBorder=Turbine.UI.Control();
	self.PreviewBorder:SetParent(self);
	self.PreviewBorder:SetSize(self:GetWidth()-42-self.PreviewCaption:GetWidth(),62);
	self.PreviewBorder:SetPosition(self.PreviewCaption:GetLeft()+self.PreviewCaption:GetWidth()+2,self.PreviewCaption:GetTop())
	self.PreviewBorder:SetBackColor(trimColor);
	self.Preview=Turbine.UI.Control();
	self.Preview:SetParent(self.PreviewBorder);
	self.Preview:SetPosition(1,1);
	self.Preview:SetSize(self:GetWidth()-44-self.PreviewCaption:GetWidth(),60);
	self.Preview:SetBackColor(backColor);
	self.Preview.Clips={};

	self:SetWantsKeyEvents(true);
	self:SetWantsUpdates(true)

	self.SlotList:SetZOrder(self.Preview:GetZOrder()+1)
	self.ChannelList:SetZOrder(self.SlotList:GetZOrder()+1)


	self.RemoveShellCommands=function(sender,args)
		Turbine.Shell.RemoveCommand(self.shellCmd)
	end
	self.shellCmd=Turbine.ShellCommand();
	local numberOfCommandsRegistered = Turbine.Shell.AddCommand("Rainbow",self.shellCmd);

	self.shellCmd.Execute = function(sender, cmd, args)
		if (string.lower(cmd) == "rainbow") then
			if (string.lower(args)=="show") then
				qsWindow:SetVisible(true);
			elseif (string.lower(args)=="hide") then
				self:SetVisible(false);
				qsWindow:SetVisible(false);
			elseif (string.lower(args)=="toggle") then
				qsWindow:SetVisible(not qsWindow:IsVisible());
			elseif (string.lower(args)=="setup") then
				qsWindow:SetVisible(true);
				mainWindow:SetVisible(true);
			else
				Turbine.Shell.WriteLine("usage:");
				Turbine.Shell.WriteLine("  /rainbow show");
				Turbine.Shell.WriteLine("       Displays the Rainbow Chat QuickSlot bar.");
				Turbine.Shell.WriteLine("  /rainbow hide");
				Turbine.Shell.WriteLine("       Hides the Rainbow Chat QuickSlot bar.");
				Turbine.Shell.WriteLine("  /rainbow toggle");
				Turbine.Shell.WriteLine("       Toggles the Rainbow Chat QuickSlot bar (shows it if it was hidden, hides it if it was showing).");
				Turbine.Shell.WriteLine("  /rainbow setup");
				Turbine.Shell.WriteLine("       Shows the Rainbow Chat Setup window.");
			end
		end
	end
	self.KeyDown=function(sender, args)
		if ( args.Action == Turbine.UI.Lotro.Action.Escape ) then
			mainWindow:SetVisible( false );
			optWindow:SetVisible( false );
		end
		if ( args.Action == 268435635 ) then -- toggle HUD
			mainWindow:SetVisible( false );
			qsWindow:SetVisible( not qsWindow:IsVisible() );
			optWindow:SetVisible( false );
		end
	end

	self.shellCmd.GetHelp = function(sender, cmd)
		return("usage:\n  /rainbow show\n       Displays the Rainbow Chat QuickSlot bar.\n  /rainbow hide\n       Hides the Rainbow chat QuickSlot bar.\n  /rainbow toggle\n       Toggles the Rainbow Chat QuickSlot bar (shows it if it was hidden, hides it if it was showing).\n  /rainbow setup\n       Shows the Rainbow Chat Setup window.");
	end

	self.shellCmd.GetShortHelp = function(sender, cmd)
		return("Rainbow - usage: /rainbow show | hide | toggle | setup");
	end
	self.Closing=function()
	end

	self.GetShowIconWhenMinimized=function()
		return self.ShowIconWhenMinimized;
	end

	self.SetTrimColor=function()
		self.Color:SetTrimColor(trimColor);
		self.Sep1:SetBackColor(trimColor);
		self.Sep2:SetBackColor(trimColor);
		self.Sep3:SetBackColor(trimColor);
		self.PreviewBorder:SetBackColor(trimColor);
	end
	self.SetBackColor2=function()
--		self.TextCaption:SetBackColor(backColor);
		self.TextEntry:SetBackColor(backColor);
--		self.ColorCaption:SetBackColor(backColor);
		self.Color:SetBackColor(backColor);
--		self.ChannelCaption:SetBackColor(backColor);
		self.ChannelList:SetBackColor(backColor);
		self.ChannelList:SetCurrentBackColor(backColor);
--		self.SlotCaption:SetBackColor(backColor);
		self.SlotList:SetBackColor(backColor);
		self.SlotList:SetCurrentBackColor(backColor);
		self.Preview:SetBackColor(backColor);
		self.ChannelOther:SetBackColor(backColor);
	end
	self.SetFontColor=function()
		self.TextCaption:SetForeColor(fontColor);
		self.TextEntry:SetForeColor(fontColor);
		self.ColorCaption:SetForeColor(fontColor);
		self.Color:SetColor(fontColor);
		self.CBNoColor:SetForeColor(fontColor);
		self.ClearButton:SetForeColor(fontColor);
		self.AddButton:SetForeColor(fontColor);
		self.ReplaceButton:SetForeColor(fontColor);
		self.ChannelCaption:SetForeColor(fontColor);
		self.ChannelList:SetTextColor(fontColor);
		self.QSButton:SetForeColor(fontColor);
		self.SlotCaption:SetForeColor(fontColor);
		self.SlotList:SetTextColor(fontColor);
		self.AssignButton:SetForeColor(fontColor);
		self.RemoveButton:SetForeColor(fontColor);
		self.PreviewCaption:SetForeColor(fontColor);
		self.ChannelOther:SetForeColor(fontColor);
	end
	self.SetLanguage=function()
		self:SetText(Resource[language][2]);
		self.TextCaption:SetText(Resource[language][24]..":");
		self.ColorCaption:SetText(Resource[language][22]..":");
		self.ClearButton:SetText(Resource[language][16]);
		self.AddButton:SetText(Resource[language][15]);
		self.ReplaceButton:SetText(Resource[language][19]);
		self.ChannelCaption:SetText(Resource[language][20]..":");
		self.QSButton:SetText(Resource[language][17]);
		self.SlotCaption:SetText(Resource[language][21]..":");
		self.AssignButton:SetText(Resource[language][18]);
		self.RemoveButton:SetText(Resource[language][23]);
		self.PreviewCaption:SetText(Resource[language][14]..":");
		self.CBNoColor:SetText(Resource[language][25]);
		self.RecallButton:SetText(Resource[language][26]);
	end

	self.Update=function( sender, args )
		if (Plugins["Rainbow"] ~= nil) and (not self.Loaded) then
			self.Loaded=true;
			Turbine.Shell.WriteLine("Rainbow Chat "..Plugins["Rainbow"]:GetVersion()..Resource[language][13]);
			Plugins["Rainbow"].Unload = function(self,sender,args)
				UnloadPlugin();
			end
		end
	end
	self.AddText=function(sender,text,color,noColor)

		local words=string.split(":"..text," "); -- we concatenate a colon to preserve leading spaces (the split function has an issue with leading spaces)
		local tmpClipIndex;
		local tmpWordIndex=1;
		local tmpString="";
		local tmpLen=0;
		local maxWidth=self.Preview:GetWidth()-1;
		if #words>0 then
			words[1]=string.sub(words[1],2);
		end
		while tmpWordIndex<=#words do
			tmpClipIndex=#self.Preview.Clips+1;
			self.Preview.Clips[tmpClipIndex]=Turbine.UI.Label();
			self.Preview.Clips[tmpClipIndex]:SetParent(self.Preview);
			self.Preview.Clips[tmpClipIndex]:SetFont(Turbine.UI.Lotro.Font.Verdana12);
			self.Preview.Clips[tmpClipIndex]:SetHeight(12);
			self.Preview.Clips[tmpClipIndex].width=0;
			self.Preview.Clips[tmpClipIndex]:SetMultiline(false);
			self.Preview.Clips[tmpClipIndex].Index=tmpClipIndex;
			self.Preview.Clips[tmpClipIndex].NoColor=noColor;
			if noColor then
				self.Preview.Clips[tmpClipIndex]:SetForeColor(fontColor);
			else
				self.Preview.Clips[tmpClipIndex]:SetForeColor(color);
			end
			self.Preview.Clips[tmpClipIndex].MouseClick=function(sender)
				if self.SelectedClip~=0 then
					self.Preview.Clips[self.SelectedClip]:SetBackColor(backColor);
				end
				self.SelectedClip=sender.Index;
				self.Preview.Clips[self.SelectedClip]:SetBackColor(Turbine.UI.Color(1,1,1));
				if sender.NoColor then
					self.Color:SetColor(fontColor);
				else
					self.Color:SetColor(self.Preview.Clips[self.SelectedClip]:GetForeColor());
				end
				self.TextEntry:SetText(self.Preview.Clips[self.SelectedClip]:GetText());
				self.CBNoColor:SetChecked(self.Preview.Clips[self.SelectedClip].NoColor);

			end

			if tmpClipIndex==1 then
				self.Preview.Clips[tmpClipIndex]:SetLeft(0);
				self.Preview.Clips[tmpClipIndex]:SetLeft(0);
			else
				self.Preview.Clips[tmpClipIndex]:SetLeft(self.Preview.Clips[tmpClipIndex-1]:GetLeft()+self.Preview.Clips[tmpClipIndex-1].width);
				self.Preview.Clips[tmpClipIndex]:SetTop(self.Preview.Clips[tmpClipIndex-1]:GetTop());
			end

			tmpString=words[tmpWordIndex];

			-- for some reason, the fontMetric doesn't always get a chance to update with multiple values so the only way to currently
			-- guarantee accuracy is to recreate the control for each request. horribly inefficient, but I don't have time to fix it atm
			self.fontMetric=nil;
			self.fontMetric=FontMetric();
			self.fontMetric:SetFont(Turbine.UI.Lotro.Font.Verdana12); -- use the default chat font
			tmpLen=self.fontMetric:GetTextWidth(tmpString)

			self.Preview.Clips[tmpClipIndex]:SetText(tmpString);
			self.Preview.Clips[tmpClipIndex]:SetWidth(tmpLen);
			self.Preview.Clips[tmpClipIndex].width=tmpLen;

			if (tmpWordIndex>1 and self.Preview.Clips[tmpClipIndex]:GetLeft()>0) or (self.Preview.Clips[tmpClipIndex]:GetLeft()+tmpLen>=maxWidth) then
				-- first word didn't fit, need new line
				if self.Preview.Clips[tmpClipIndex]:GetLeft()>0 then
					self.Preview.Clips[tmpClipIndex]:SetText(words[tmpWordIndex]); -- eliminate the space we added earlier
					self.Preview.Clips[tmpClipIndex]:SetLeft(0);
					self.Preview.Clips[tmpClipIndex]:SetTop(self.Preview.Clips[tmpClipIndex]:GetTop()+14);
				end
			end

			tmpWordIndex=tmpWordIndex+1;
			if tmpWordIndex<=#words and self.Preview.Clips[tmpClipIndex]:GetLeft()+tmpLen<maxWidth then
				-- try additional words until we exceed the width
				tmpString=self.Preview.Clips[tmpClipIndex]:GetText().." "..words[tmpWordIndex];

				self.fontMetric=nil;
				self.fontMetric=FontMetric();
				self.fontMetric:SetFont(Turbine.UI.Lotro.Font.Verdana12); -- use the default chat font
				tmpLen=self.fontMetric:GetTextWidth(tmpString);

				while (tmpWordIndex<=#words) and ((self.Preview.Clips[tmpClipIndex]:GetLeft()+tmpLen)<=maxWidth) do
					self.Preview.Clips[tmpClipIndex]:SetText(tmpString);
					self.Preview.Clips[tmpClipIndex]:SetWidth(tmpLen);
					self.Preview.Clips[tmpClipIndex].width=tmpLen;
					tmpWordIndex=tmpWordIndex+1;
					if tmpWordIndex<=#words then
						tmpString=self.Preview.Clips[tmpClipIndex]:GetText().." "..words[tmpWordIndex];

						self.fontMetric=nil;
						self.fontMetric=FontMetric();
						self.fontMetric:SetFont(Turbine.UI.Lotro.Font.Verdana12); -- use the default chat font
						tmpLen=self.fontMetric:GetTextWidth(tmpString);
					end
				end
			end

			tmpClipIndex=tmpClipIndex+1;
		end		
		-- now add the text to the SayText and update the quickslot
		color.RHex=string.format("%x",255*(color.R))
		while string.len(color.RHex)<2 do
			color.RHex="0"..color.RHex;
		end
		color.GHex=string.format("%x",255*(color.G))
		while string.len(color.GHex)<2 do
			color.GHex="0"..color.GHex;
		end
		color.BHex=string.format("%x",255*(color.B))
		while string.len(color.BHex)<2 do
			color.BHex="0"..color.BHex;
		end


		if self.CBNoColor:IsChecked() then
			self.SayText=self.SayText..text;
		else
			self.SayText=self.SayText.."<rgb=#"..color.RHex..color.GHex..color.BHex..">"..text.."</rgb>";
		end
		self:SetQS();
	end
	self.SetQS=function()
		if self.SayText=="" then
			self.QS:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, ""));
		else
			if self.ChannelOther:IsVisible() then
				self.QS:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self.ChannelOther:GetText().." "..self.SayText));
			else
				self.QS:SetShortcut(Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self.ChannelList:GetValue().." "..self.SayText));
			end
		end
	end
	self.ReplaceText=function()
		-- currently, we just brute force reassign the text in the field, size it and rebuild the SayText string
		-- no attempt is made to word wrap the preview on a replace operation
		self.Preview.Clips[self.SelectedClip]:SetText(self.TextEntry:GetText());
		self.Preview.Clips[self.SelectedClip].NoColor=self.CBNoColor:IsChecked();
		if self.CBNoColor:IsChecked() then
			self.Preview.Clips[self.SelectedClip]:SetForeColor(fontColor);
		else
			self.Preview.Clips[self.SelectedClip]:SetForeColor(self.Color:GetColor());
		end

		-- now re-build the SayText string starting from scratch :(
		self.SayText="";
		local tmpIndex;
		local color;
		for tmpIndex=1,#self.Preview.Clips do
			color=self.Preview.Clips[tmpIndex]:GetForeColor();
			color.RHex=string.format("%x",255*(color.R))
			while string.len(color.RHex)<2 do
				color.RHex="0"..color.RHex;
			end
			color.GHex=string.format("%x",255*(color.G))
			while string.len(color.GHex)<2 do
				color.GHex="0"..color.GHex;
			end
			color.BHex=string.format("%x",255*(color.B))
			while string.len(color.BHex)<2 do
				color.BHex="0"..color.BHex;
			end

			if self.Preview.Clips[tmpIndex].NoColor then
				self.SayText=self.SayText..self.Preview.Clips[tmpIndex]:GetText();
			else
				self.SayText=self.SayText.."<rgb=#"..color.RHex..color.GHex..color.BHex..">"..self.Preview.Clips[tmpIndex]:GetText().."</rgb>";
			end
		end
		self:SetQS();
	end
end

qsWindow=QuickSlotWindow();
LoadQuickSlots();
mainWindow=RainbowWindow();
optWindow=OptionsWindow();
mainWindow:SetZOrder(mainWindow:GetZOrder()+1);
optWindow:SetZOrder(mainWindow:GetZOrder()+1);
qsWindow:SetVisible(true);

